import math

def ontheirfeetBehavior(bubble,context):
    inbubble = context
    if 'gravity' in context.getTag():
        dx,dy = bubble.getDelta(inbubble)
        dist = bubble.getDistSQ(inbubble)**0.5
        bubble = bubble.setDirection(dx/dist,dy/dist)
    return bubble

def movableBehavior(bubble, context):
    bubbleWill = bubble.giggle()
    inbubble = context.recursivePredicate(bubbleWill)
    if 'solid' in inbubble.getTag():
        dx,dy = inbubble.getDelta(bubble)
        dist = inbubble.getDistSQ(bubble)**0.5
        return bubble.stop()#.push(-0.2*dx/dist,-0.2*dy/dist)
    return bubbleWill

def fallingBehavior(bubble,context):
    constant = 0.006
    inbubble = context
    if 'gravity' in inbubble.getTag():
        deltaX,deltaY = bubble.getDelta(inbubble)
        dist = bubble.getDistSQ(inbubble)**0.5
        return bubble.push(-constant*deltaX/dist,-constant*deltaY/dist)
    return bubble

def springBehavior(length,force,forceLink):
    def closure(bubble,context):
        fx,fy = 0,0
        newLink = []
        for link in bubble.getLink():
            dx,dy = bubble.getDelta(link)
            dist = math.hypot(dx,dy)
            vx = dx/dist if dist !=0 else dx
            vy = dy/dist if dist !=0 else dy
            diff = (length-abs(dist))
            fx += force*diff*vx
            fy += force*diff*vy
            newLink.append(link.push(-forceLink*diff*vx,-forceLink*diff*vy))
        return bubble.push(fx,fy).updateLink(newLink)
    return closure

def destroyableBehavior(bubble,context):
    link = list(bubble.getLink())
    if 'bullet' in context.getTag():
        x,y = context.getPosition()
        explosion = Bubble('hole',\
                            context.getPos(),\
                            10,\
                            (0,-1),\
                            (0,0),\
                            [],\
                            [],\
                            AinsideB,\
                            [explodingBehavior],\
                            ['explosion']).translate(x,y)
        link.append(explosion)
    return bubble.updateLink(link)

def explodingBehavior(bubble,context):
    if not 'bullet' in context.getTag():
        return bubble.removeTag('explosion')
    return bubble
    
def bulletBehavior(bubble,context):
    if 'explosion' in context.getTag():
        return bubble.addTag('vanish')
    return bubble

def vanishableLinkBehavior(bubble,context):
    newLnks = []
    for link in bubble.getLink():
        if not 'vanish' in link.getTag():
            newLnks.append(link)
    return bubble.updateLink(newLnks)

suite = {
    'movableBehavior':movableBehavior,
    'fallingBehavior':fallingBehavior,
    'springBehavior':springBehavior(20,0.02,0.08),
    'ontheirfeetBehavior':ontheirfeetBehavior,
    
    'destroyableBehavior':destroyableBehavior,
    'explodingBehavior':explodingBehavior,
    'bulletBehavior':bulletBehavior,
    
    'vanishableLinkBehavior':vanishableLinkBehavior
}
