import math
import behavior
import effect
import predicate
import logging

import defensive
import obj

test = defensive.Testing()

@obj.immutable
class Bubble(object):
    def __init__(self,name,\
                 (x,y),r,\
                 (dirx,diry),
                 (vx,vy),\
                 child,\
                 link,\
                 predicate,\
                 behavior,\
                 tag):
        self._name = name
        self._position= (x,y)
        self._r = r
        self._direction = (dirx,diry)
        self._forceVector = (vx,vy)
        self._child = list(child)
        self._link = list(link)
        self._predicate = predicate
        self._behavior = list(behavior)
        self._tag = list(tag)
    
    def getName(self):
        """
        return name, simple string
        """
        return self._name
    
    def getPosition(self):
        """
        return x,y position
        """
        return self._position    
    def getNextPos(self):
        """
        return predicted x,y position
        """
        x,y = self._position
        vx,vy = self._forceVector
        return (x+vx,y+vy)
    
    def getR(self):
        """
        return radius
        """
        return self._r

    def getDirection(self):
        """
        return object orientation, not the same as real movement
        """
        return self._direction

    def getForceVector(self):
        """
        return the force direction
        """
        return self._forceVector
    
    def getPredicate(self):
        return self._predicate
    
    def getSolid(self):
        return self._solid

    def _getChild(self):
        for child in self._child:
            yield child
      
    def getChild(self):
        for child in self._child:
            x,y = self._position
            yield child.translate(x, y)
    
    def getLink(self):
        for link in self._link:
            yield link
            
    def getBehavior(self):
        for behavior in self._behavior:
            yield behavior
    
    def getTag(self):
        for tag in self._tag:
            yield tag

    def getRadian(self):
        dirx,diry = self._direction
        return math.atan2(dirx,diry)

    def getDelta(self,other):
        """
        return delta of repective (x,y) position
        """
        ax,ay = self._position
        bx,by = other._position
        deltaX = ax - bx
        deltaY = ay - by
        return deltaX,deltaY
    
    def getDistSQ(self,other):
        """
        return square of distance between two bubble
        """
        deltaX,deltaY = self.getDelta(other)
        return deltaX**2 + deltaY**2
    
    def intersect(self,other):
        """
        return true if two bubble intersect
        """
        minDist = self.getR() + other.getR()
        minDistSQ = minDist**2
        return minDistSQ >= self.getDistSQ(other)
    
    def inside(self,other):
        """
        return true if self is completely inside the other
        """
        minDist = other.getR() - self.getR()
        minDistSQ = minDist**2
        return minDistSQ >= self.getDistSQ(other)

    def _forceLimit(self):
        vx,vy = self._forceVector
        if vx > 1:
            vx = 1
        if vx < -1:
            vx = -1
        if vy > 1:
            vy = 1
        if vy < -1:
            vy = -1
        return Bubble(self._name,\
                      self._position,\
                      self._r,\
                      self._direction,\
                      (vx,vy),\
                      self._child,\
                      self._link,\
                      self._predicate,\
                      self._behavior,\
                      self._tag)

     

    def stop(self):
        return Bubble(self._name,\
                      self._position,\
                      self._r,\
                      self._direction,\
                      (0,0),\
                      self._child,\
                      self._link,\
                      self._predicate,\
                      self._behavior,\
                      self._tag)
    
    def translate(self,x,y):
        ax,ay = self._position
        return Bubble(self._name,\
                      (ax+x,ay+y),\
                      self._r,\
                      self._direction,\
                      self._forceVector,\
                      self._child,\
                      self._link,\
                      self._predicate,\
                      self._behavior,\
                      self._tag)
    
    def rotate(self,rad):
        s = math.sin(rad)
        c = math.cos(rad)
        x,y = self._direction
        dirx = x*c - y*s
        diry = x*s +y*c
        dist = math.hypot(dirx,diry)
        dirx /= dist
        diry /= dist
        return Bubble(self._name,\
                      self.
                      _position,\
                      self._r,\
                      (dirx,diry),\
                      self._forceVector,\
                      self._child,\
                      self._link,\
                      self._predicate,\
                      self._behavior,\
                      self._tag)
                      
    def setDirection(self,dirx,diry):
        return Bubble(self._name,\
                      self._position,\
                      self._r,\
                      (dirx,diry),\
                      self._forceVector,\
                      self._child,\
                      self._link,\
                      self._predicate,\
                      self._behavior,\
                      self._tag)
    
    def push(self,x,y):
        vx,vy = self._forceVector
        vx = vx+x
        vy = vy+y
        return Bubble(self._name,\
                      self._position,\
                      self._r,\
                      self._direction,\
                      (vx,vy),\
                      self._child,\
                      self._link,\
                      self._predicate,\
                      self._behavior,\
                      self._tag)._forceLimit()

    def restrain(self,scale):
        vx,vy = self._forceVector
        vx = vx*scale
        vy = vy*scale
        return Bubble(self._name,\
                      self._position,\
                      self._r,\
                      self._direction,\
                      (vx,vy),\
                      self._child,\
                      self._link,\
                      self._predicate,\
                      self._behavior,\
                      self._tag)._forceLimit()

    def giggle(self):
        x,y = self._position
        vx, vy = self._forceVector
        return Bubble(self._name,\
                      (x+vx,y+vy),\
                      self._r,\
                      self._direction,\
                      (vx,vy),\
                      self._child,
                      self._link,\
                      self._predicate,\
                      self._behavior,\
                      self._tag)._forceLimit()

    def updateChild(self,child):
        return Bubble(self._name,\
                      self._position,\
                      self._r,\
                      self._direction,\
                      self._forceVector,\
                      list(child),\
                      self._link,\
                      self._predicate,\
                      self._behavior,\
                      self._tag)
    
    def updateLink(self,newlink):
        return Bubble(self._name,\
                      self._position,\
                      self._r,\
                      self._direction,\
                      self._forceVector,\
                      self._child,\
                      newlink,\
                      self._predicate,\
                      self._behavior,\
                      self._tag)

    def addTag(self,newTag):
        return Bubble(self._name,\
                      self._position,\
                      self._r,\
                      self._direction,\
                      self._forceVector,\
                      self._child,\
                      self._link,\
                      self._predicate,\
                      self._behavior,\
                      [newTag]+self._tag)

    def addTag(self,rmTag):
        return Bubble(self._name,\
                      self._position,\
                      self._r,\
                      self._direction,\
                      self._forceVector,\
                      self._child,\
                      self._link,\
                      self._predicate,\
                      self._behavior,\
                      list(self._tag).remove(rmTag))                    
                      
    def recursivePredicate(self,bubble,cnt=1):
        if self.getPredicate()(bubble,self):
            for child in self.getChild():
                if child.getPredicate()(bubble,child):
                    return child.recursivePredicate(bubble,cnt+1)
        return self

    def against(self,context,cnt=1):
        inbubble = context.recursivePredicate(self)
        newSelf = self
        for behavior in newSelf.getBehavior():
            newSelf = behavior(newSelf,inbubble)
        newLink=[]
        for link in newSelf.getLink():
            newLink.append(link.against(context,cnt+1))
        return newSelf.updateLink(newLink)

#some function
def bubbleFromStruct(db,effectDict,behaviorDict,predicateDict):
    cluster = []
    for name in db.keys():
        cluster.append(Bubble(name,\
                              (db[name]['x'],db[name]['y']),\
                               db[name]['r'],\
                              (db[name]['dirx'],db[name]['diry']),\
                              (db[name]['vx'],db[name]['vy']),\
                               bubbleFromStruct(db[name]['child'],effectDict,behaviorDict,predicateDict),\
                               bubbleFromStruct(db[name]['link'],effectDict,behaviorDict,predicateDict),\
                               predicateDict[db[name]['predicate']],\
                               (behaviorDict[behavior] for behavior in db[name]['behavior']),\
                               db[name]['tag']))
    return cluster

if __name__ == '__main__':
    pass
